@@ -77,6 +77,7 @@ static_assert(_XKB_MOD_INDEX_NUM_ENTRIES == 8, "Invalid X11 core modifiers");
 
 enum xkb_action_type {
     ACTION_TYPE_NONE = 0,
+    ACTION_TYPE_VOID, /* libxkbcommon extension */
     ACTION_TYPE_MOD_SET,
     ACTION_TYPE_MOD_LATCH,
     ACTION_TYPE_MOD_LOCK,
@@ -91,7 +92,9 @@ enum xkb_action_type {
     ACTION_TYPE_SWITCH_VT,
     ACTION_TYPE_CTRL_SET,
     ACTION_TYPE_CTRL_LOCK,
+    ACTION_TYPE_UNSUPPORTED_LEGACY,
     ACTION_TYPE_PRIVATE,
+    ACTION_TYPE_INTERNAL, /* Action specific and internal to xkbcommon */
     _ACTION_TYPE_NUM_ENTRIES
 };
 
@@ -189,6 +192,21 @@ struct xkb_private_action {
     uint8_t data[7];
 };
 
+enum xkb_internal_action_flags {
+    INTERNAL_BREAKS_GROUP_LATCH = (1 << 0),
+    INTERNAL_BREAKS_MOD_LATCH = (1 << 1),
+};
+
+/* Action specific and internal to xkbcommon */
+struct xkb_internal_action {
+    enum xkb_action_type type;
+    enum xkb_internal_action_flags flags;
+    union {
+        /* flag == INTERNAL_BREAKS_MOD_LATCH */
+        xkb_mod_mask_t clear_latched_mods;
+    };
+};
+
 union xkb_action {
     enum xkb_action_type type;
     struct xkb_mod_action mods;
@@ -199,6 +217,7 @@ union xkb_action {
     struct xkb_pointer_action ptr;
     struct xkb_pointer_button_action btn;
     struct xkb_private_action priv;
+    struct xkb_internal_action internal;
 };
 
 struct xkb_key_type_entry {
@@ -211,9 +230,9 @@ struct xkb_key_type {
     xkb_atom_t name;
     struct xkb_mods mods;
     xkb_level_index_t num_levels;
-    unsigned int num_level_names;
+    xkb_level_index_t num_level_names;
     xkb_atom_t *level_names;
-    unsigned int num_entries;
+    darray_size_t num_entries;
     struct xkb_key_type_entry *entries;
 };
 
@@ -361,7 +380,7 @@ struct xkb_mod {
 
 struct xkb_mod_set {
     struct xkb_mod mods[XKB_MAX_MODS];
-    unsigned int num_mods;
+    xkb_mod_index_t num_mods;
     xkb_mod_mask_t explicit_vmods;
 };
 
@@ -410,16 +429,46 @@ struct xkb_keymap {
     struct xkb_key *keys;
 
     /* aliases in no particular order */
-    unsigned int num_key_aliases;
+    darray_size_t num_key_aliases;
     struct xkb_key_alias *key_aliases;
 
     struct xkb_key_type *types;
-    unsigned int num_types;
+    darray_size_t num_types;
 
-    unsigned int num_sym_interprets;
+    darray_size_t num_sym_interprets;
     struct xkb_sym_interpret *sym_interprets;
 
+    /**
+     * Modifiers configuration.
+     * This is *internal* to the keymap; other implementations may use different
+     * virtual modifiers indexes. Ours depends on:
+     *   1. the order of the parsing of the keymap components;
+     *   2. the order of the virtual modifiers declarations;
+     */
     struct xkb_mod_set mods;
+    /**
+     * Modifier mask of the *canonical* state, i.e. the mask with the *smallest*
+     * population count that denotes all bits used to encode the modifiers in
+     * the keyboard state. It is equal to the bitwise OR of *real* modifiers and
+     * all *virtual* modifiers mappings.
+     *
+     * [WARNING] The bits that do not correspond to *real* modifiers should
+     * *not* be interpreted as corresponding to indexes of virtual modifiers of
+     * the keymap. Indeed, one may use explicit vmod mapping with an arbitrary
+     * value.
+     *
+     * E.g. if M1 is the only vmod and it is defined by:
+     *
+     *     virtual_modifiers M1=0x80000000; // 1 << (32 - 1)
+     *
+     * then the 32th bit of a modifier mask input does *not* denote the 32th
+     * virtual modifier of the keymap, but merely the encoding of the mapping of
+     * M1.
+     *
+     * In the API, any input mask should be preprocessed to resolve the bits
+     * that do not match the canonical mask.
+     */
+    xkb_mod_mask_t canonical_state_mask;
 
     /* This field has 2 uses:
      * • During parsing: Expected layouts count after RMLVO resolution, if any;
@@ -430,7 +479,7 @@ struct xkb_keymap {
     xkb_atom_t *group_names;
 
     struct xkb_led leds[XKB_MAX_LEDS];
-    unsigned int num_leds;
+    xkb_led_index_t num_leds;
 
     char *keycodes_section_name;
     char *symbols_section_name;
@@ -546,16 +595,16 @@ XkbWrapGroupIntoRange(int32_t group,
                       enum xkb_range_exceed_type out_of_range_group_action,
                       xkb_layout_index_t out_of_range_group_number);
 
-xkb_mod_mask_t
+XKB_EXPORT_PRIVATE xkb_mod_mask_t
 mod_mask_get_effective(struct xkb_keymap *keymap, xkb_mod_mask_t mods);
 
 struct xkb_level *
 xkb_keymap_key_get_level(struct xkb_keymap *keymap, const struct xkb_key *key,
                          xkb_layout_index_t layout, xkb_level_index_t level);
 
-unsigned int
+xkb_action_count_t
 xkb_keymap_key_get_actions_by_level(struct xkb_keymap *keymap,
-                                    xkb_keycode_t kc,
+                                    const struct xkb_key *key,
                                     xkb_layout_index_t layout,
                                     xkb_level_index_t level,
                                     const union xkb_action **actions);
