@@ -729,7 +729,7 @@ parse_config_item(struct rxkb_context *ctx, xmlNode *parent,
                     config->popularity = RXKB_POPULARITY_EXOTIC;
                 else
                     log_err(ctx, XKB_LOG_MESSAGE_NO_ID,
-                            "xml:%d: invalid popularity attribute: expected "
+                            "xml:%u: invalid popularity attribute: expected "
                             "'standard' or 'exotic', got: '%s'\n",
                             ci->line, raw_popularity);
             }
@@ -751,7 +751,7 @@ parse_config_item(struct rxkb_context *ctx, xmlNode *parent,
 
             if (!config->name || !strlen(config->name))  {
                 log_err(ctx, XKB_LOG_MESSAGE_NO_ID,
-                        "xml:%d: missing required element 'name'\n",
+                        "xml:%u: missing required element 'name'\n",
                         ci->line);
                 config_item_free(config);
                 return false;
@@ -1196,7 +1196,7 @@ validate(struct rxkb_context *ctx, xmlDoc *doc)
     xmlCtxtSetOptions(xmlCtxt, _XML_OPTIONS | XML_PARSE_DTDLOAD);
 
     xmlParserInputPtr pinput =
-        xmlNewInputFromMemory(NULL, dtdstr, sizeof(dtdstr),
+        xmlNewInputFromMemory(NULL, dtdstr, ARRAY_SIZE(dtdstr) - 1,
                               XML_INPUT_BUF_STATIC);
     if (!pinput)
         goto dtd_error;
@@ -1206,11 +1206,11 @@ validate(struct rxkb_context *ctx, xmlDoc *doc)
     /* Note: do not use xmlParserInputBufferCreateStatic, it generates random
      * DTD validity errors for unknown reasons */
     xmlParserInputBufferPtr buf =
-        xmlParserInputBufferCreateMem(dtdstr, sizeof(dtdstr),
+        xmlParserInputBufferCreateMem(dtdstr, ARRAY_SIZE(dtdstr) - 1,
                                       XML_CHAR_ENCODING_NONE);
     if (!buf)
         goto dtd_error;
-    xmlDtd *dtd = dtd = xmlIOParseDTD(NULL, buf, XML_CHAR_ENCODING_UTF8);
+    xmlDtd *dtd = xmlIOParseDTD(NULL, buf, XML_CHAR_ENCODING_UTF8);
 #endif
 
     if (!dtd) {
