@@ -138,6 +138,7 @@ const LookupEntry useModMapValueNames[] = {
 
 const LookupEntry actionTypeNames[] = {
     { "NoAction", ACTION_TYPE_NONE },
+    { "VoidAction", ACTION_TYPE_VOID },
     { "SetMods", ACTION_TYPE_MOD_SET },
     { "LatchMods", ACTION_TYPE_MOD_LATCH },
     { "LockMods", ACTION_TYPE_MOD_LOCK },
@@ -161,24 +162,24 @@ const LookupEntry actionTypeNames[] = {
     { "LockControls", ACTION_TYPE_CTRL_LOCK },
     { "Private", ACTION_TYPE_PRIVATE },
     /* deprecated actions below here - unused */
-    { "RedirectKey", ACTION_TYPE_NONE },
-    { "Redirect", ACTION_TYPE_NONE },
-    { "ISOLock", ACTION_TYPE_NONE },
-    { "ActionMessage", ACTION_TYPE_NONE },
-    { "MessageAction", ACTION_TYPE_NONE },
-    { "Message", ACTION_TYPE_NONE },
-    { "DeviceBtn", ACTION_TYPE_NONE },
-    { "DevBtn", ACTION_TYPE_NONE },
-    { "DevButton", ACTION_TYPE_NONE },
-    { "DeviceButton", ACTION_TYPE_NONE },
-    { "LockDeviceBtn", ACTION_TYPE_NONE },
-    { "LockDevBtn", ACTION_TYPE_NONE },
-    { "LockDevButton", ACTION_TYPE_NONE },
-    { "LockDeviceButton", ACTION_TYPE_NONE },
-    { "DeviceValuator", ACTION_TYPE_NONE },
-    { "DevVal", ACTION_TYPE_NONE },
-    { "DeviceVal", ACTION_TYPE_NONE },
-    { "DevValuator", ACTION_TYPE_NONE },
+    { "RedirectKey", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "Redirect", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "ISOLock", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "ActionMessage", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "MessageAction", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "Message", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "DeviceBtn", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "DevBtn", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "DevButton", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "DeviceButton", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "LockDeviceBtn", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "LockDevBtn", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "LockDevButton", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "LockDeviceButton", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "DeviceValuator", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "DevVal", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "DeviceVal", ACTION_TYPE_UNSUPPORTED_LEGACY },
+    { "DevValuator", ACTION_TYPE_UNSUPPORTED_LEGACY },
     { NULL, 0 },
 };
 
@@ -291,7 +292,7 @@ LedStateMaskText(struct xkb_context *ctx, const LookupEntry *lookup,
     if (mask == 0)
         return "0";
 
-    for (unsigned i = 0; mask; i++) {
+    for (unsigned int i = 0; mask; i++) {
         int ret;
 
         if (!(mask & (1u << i)))
@@ -325,7 +326,7 @@ ControlMaskText(struct xkb_context *ctx, enum xkb_action_controls mask)
     if (mask == CONTROL_ALL)
         return "all";
 
-    for (unsigned i = 0; mask; i++) {
+    for (unsigned int i = 0; mask; i++) {
         int ret;
 
         if (!(mask & (1u << i)))
