@@ -14,9 +14,13 @@
 
 #include "config.h"
 
+#include "context.h"
+#include "keymap.h"
+#include "messages-codes.h"
 #include "xkbcomp-priv.h"
 #include "text.h"
 #include "expr.h"
+#include "xkbcomp/ast.h"
 #include "action.h"
 
 static const ExprBoolean constTrue = {
@@ -55,17 +59,10 @@ enum action_field {
     ACTION_FIELD_MODS_TO_CLEAR,
 };
 
-ActionsInfo *
-NewActionsInfo(void)
+void
+InitActionsInfo(ActionsInfo *info)
 {
-    enum xkb_action_type type;
-    ActionsInfo *info;
-
-    info = calloc(1, sizeof(*info));
-    if (!info)
-        return NULL;
-
-    for (type = 0; type < _ACTION_TYPE_NUM_ENTRIES; type++)
+    for (enum xkb_action_type type = 0; type < _ACTION_TYPE_NUM_ENTRIES; type++)
         info->actions[type].type = type;
 
     /* Apply some "factory defaults". */
@@ -75,14 +72,6 @@ NewActionsInfo(void)
     info->actions[ACTION_TYPE_PTR_DEFAULT].dflt.value = 1;
     info->actions[ACTION_TYPE_PTR_MOVE].ptr.flags = ACTION_ACCEL;
     info->actions[ACTION_TYPE_SWITCH_VT].screen.flags = ACTION_SAME_SCREEN;
-
-    return info;
-}
-
-void
-FreeActionsInfo(ActionsInfo *info)
-{
-    free(info);
 }
 
 static const LookupEntry fieldStrings[] = {
@@ -164,7 +153,7 @@ static inline bool
 ReportIllegal(struct xkb_context *ctx, enum xkb_action_type action,
               enum action_field field)
 {
-    log_err(ctx, XKB_LOG_MESSAGE_NO_ID,
+    log_err(ctx, XKB_ERROR_INVALID_ACTION_FIELD,
             "Field %s is not defined for an action of type %s; "
             "Action definition ignored\n",
             fieldText(field), ActionTypeText(action));
@@ -188,7 +177,11 @@ HandleNoAction(struct xkb_context *ctx, const struct xkb_mod_set *mods,
                const ExprDef *array_ndx, const ExprDef *value)
 
 {
-    return true;
+    log_err(ctx, XKB_ERROR_INVALID_ACTION_FIELD,
+            "The \"%s\" action takes no argument, but got \"%s\" field; "
+            "Action definition ignored\n",
+            ActionTypeText(action->type), fieldText(field));
+    return false;
 }
 
 static bool
@@ -600,7 +593,8 @@ HandleSetLockControls(struct xkb_context *ctx, const struct xkb_mod_set *mods,
         act->ctrls = (enum xkb_action_controls) mask;
         return true;
     }
-    else if (field == ACTION_FIELD_AFFECT) {
+    else if (field == ACTION_FIELD_AFFECT &&
+             action->type == ACTION_TYPE_CTRL_LOCK) {
         return CheckAffectField(ctx, action->type, array_ndx, value,
                                 &act->flags);
     }
@@ -608,6 +602,16 @@ HandleSetLockControls(struct xkb_context *ctx, const struct xkb_mod_set *mods,
     return ReportIllegal(ctx, action->type, field);
 }
 
+static bool
+HandleUnsupportedLegacy(struct xkb_context *ctx, const struct xkb_mod_set *mods,
+                        union xkb_action *action, enum action_field field,
+                        const ExprDef *array_ndx, const ExprDef *value)
+
+{
+    /* Do not check any field */
+    return true;
+}
+
 static bool
 HandlePrivate(struct xkb_context *ctx, const struct xkb_mod_set *mods,
               union xkb_action *action, enum action_field field,
@@ -666,7 +670,7 @@ HandlePrivate(struct xkb_context *ctx, const struct xkb_mod_set *mods,
             size_t len = strlen(str);
             if (len < 1 || len > sizeof(act->data)) {
                 log_warn(ctx, XKB_LOG_MESSAGE_NO_ID,
-                         "A private action has %ld data bytes; "
+                         "A private action has %zu data bytes; "
                          "Illegal data ignored\n", sizeof(act->data));
                 return false;
             }
@@ -722,6 +726,7 @@ typedef bool (*actionHandler)(struct xkb_context *ctx,
 
 static const actionHandler handleAction[_ACTION_TYPE_NUM_ENTRIES] = {
     [ACTION_TYPE_NONE] = HandleNoAction,
+    [ACTION_TYPE_VOID] = HandleNoAction,
     [ACTION_TYPE_MOD_SET] = HandleSetLatchLockMods,
     [ACTION_TYPE_MOD_LATCH] = HandleSetLatchLockMods,
     [ACTION_TYPE_MOD_LOCK] = HandleSetLatchLockMods,
@@ -736,6 +741,7 @@ static const actionHandler handleAction[_ACTION_TYPE_NUM_ENTRIES] = {
     [ACTION_TYPE_SWITCH_VT] = HandleSwitchScreen,
     [ACTION_TYPE_CTRL_SET] = HandleSetLockControls,
     [ACTION_TYPE_CTRL_LOCK] = HandleSetLockControls,
+    [ACTION_TYPE_UNSUPPORTED_LEGACY] = HandleUnsupportedLegacy,
     [ACTION_TYPE_PRIVATE] = HandlePrivate,
 };
 
@@ -753,10 +759,11 @@ HandleActionDef(struct xkb_context *ctx, ActionsInfo *info,
         return false;
     }
 
-    const char *str = xkb_atom_text(ctx, def->action.name);
+    const char *action_name = xkb_atom_text(ctx, def->action.name);
     enum xkb_action_type handler_type;
-    if (!stringToActionType(str, &handler_type)) {
-        log_err(ctx, XKB_LOG_MESSAGE_NO_ID, "Unknown action %s\n", str);
+    if (!stringToActionType(action_name, &handler_type)) {
+        log_err(ctx, XKB_ERROR_UNKNOWN_ACTION_TYPE,
+                "Unknown action %s\n", action_name);
         return false;
     }
 
@@ -767,6 +774,19 @@ HandleActionDef(struct xkb_context *ctx, ActionsInfo *info,
      */
     *action = info->actions[handler_type];
 
+    if (handler_type == ACTION_TYPE_UNSUPPORTED_LEGACY) {
+        log_warn(ctx, XKB_WARNING_UNSUPPORTED_LEGACY_ACTION,
+                 "Unsupported legacy action type \"%s\".\n", action_name);
+        /*
+         * Degrade to an uneffective supported action type.
+         * Fields will still be processed with the original action type.
+         */
+        action->type = ACTION_TYPE_NONE;
+        /* FIXME: Degrading to NoAction() has the following drawbacks:
+         * - It disables overwriting a previous action.
+         * - It enables interpretations. */
+    }
+
     /*
      * Now change the action properties as specified for this
      * particular instance, e.g. "modifiers" and "clearLocks" in:
@@ -804,7 +824,7 @@ HandleActionDef(struct xkb_context *ctx, ActionsInfo *info,
 
         enum action_field fieldNdx;
         if (!stringToField(fieldRtrn, &fieldNdx)) {
-            log_err(ctx, XKB_LOG_MESSAGE_NO_ID,
+            log_err(ctx, XKB_ERROR_INVALID_ACTION_FIELD,
                     "Unknown field name %s\n", fieldRtrn);
             return false;
         }
@@ -818,21 +838,51 @@ HandleActionDef(struct xkb_context *ctx, ActionsInfo *info,
 }
 
 bool
-SetActionField(struct xkb_context *ctx, ActionsInfo *info,
-               struct xkb_mod_set *mods, const char *elem,
-               const char *field, ExprDef *array_ndx, ExprDef *value)
+SetDefaultActionField(struct xkb_context *ctx, ActionsInfo *info,
+                      struct xkb_mod_set *mods, const char *elem,
+                      const char *field, ExprDef *array_ndx,
+                      ExprDef *value, enum merge_mode merge)
 {
     enum xkb_action_type action;
-    if (!stringToActionType(elem, &action))
+    if (!stringToActionType(elem, &action)) {
+        log_err(ctx, XKB_ERROR_UNKNOWN_ACTION_TYPE, "Unknown action %s\n", elem);
         return false;
+    }
 
     enum action_field action_field;
     if (!stringToField(field, &action_field)) {
-        log_err(ctx, XKB_LOG_MESSAGE_NO_ID,
+        log_err(ctx, XKB_ERROR_INVALID_ACTION_FIELD,
                 "\"%s\" is not a legal field name\n", field);
         return false;
     }
 
-    return handleAction[action](ctx, mods, &info->actions[action],
-                                action_field, array_ndx, value);
+    union xkb_action* const into = &info->actions[action];
+    /* Initialize with current defaults to enable comparison, see aftwerwards */
+    union xkb_action from = *into;
+
+    /* Parse action */
+    if (!handleAction[action](ctx, mods, &from, action_field, array_ndx, value))
+        return false;
+
+    /*
+     * Merge action with its corresponding default
+     *
+     * NOTE: Contrary to other items, actions do not have a “defined” field, so
+     * we fallback to compare all the actions fields. The drawback is that it
+     * overmatches, as even setting an *explicit* default value for the first
+     * time (and different from the “factory” default) would *always* display a
+     * warning. So we guard the logging with a high verbosity, as best effort
+     * mitigation.
+     */
+    if (!action_equal(into, &from)) {
+        const bool replace = (merge != MERGE_AUGMENT);
+        log_vrb(ctx, 9, XKB_LOG_MESSAGE_NO_ID,
+                "Conflicting field \"%s\" for default action \"%s\"; "
+                "Using %s, ignore %s\n",
+                fieldText(action_field), ActionTypeText(action),
+                (replace ? "from" : "into"), (replace ? "into" : "from"));
+        if (replace)
+            *into = from;
+    }
+    return true;
 }
