@@ -149,6 +149,9 @@ enum rules_mlvo {
     _MLVO_NUM_ENTRIES
 };
 
+typedef uint8_t mlvo_index_t;
+typedef uint8_t mlvo_mask_t;
+
 static const struct sval rules_mlvo_svals[_MLVO_NUM_ENTRIES] = {
     [MLVO_MODEL] = SVAL_INIT("model"),
     [MLVO_LAYOUT] = SVAL_INIT("layout"),
@@ -165,6 +168,9 @@ enum rules_kccgst {
     _KCCGST_NUM_ENTRIES
 };
 
+typedef uint8_t kccgst_index_t;
+typedef uint8_t kccgst_mask_t;
+
 static const struct sval rules_kccgst_svals[_KCCGST_NUM_ENTRIES] = {
     [KCCGST_KEYCODES] = SVAL_INIT("keycodes"),
     [KCCGST_TYPES] = SVAL_INIT("types"),
@@ -198,9 +204,9 @@ struct group {
 };
 
 struct mapping {
-    int mlvo_at_pos[_MLVO_NUM_ENTRIES];
-    unsigned int num_mlvo;
-    unsigned int defined_mlvo_mask;
+    enum rules_mlvo mlvo_at_pos[_MLVO_NUM_ENTRIES];
+    mlvo_index_t num_mlvo;
+    mlvo_mask_t defined_mlvo_mask;
     bool has_layout_idx_range;
     /* This member has 2 uses:
      * • Keep track of layout and variant indexes while parsing MLVO headers.
@@ -218,9 +224,9 @@ struct mapping {
         xkb_layout_mask_t active;
         xkb_layout_mask_t layouts_candidates_mask;
     };
-    int kccgst_at_pos[_KCCGST_NUM_ENTRIES];
-    unsigned int num_kccgst;
-    unsigned int defined_kccgst_mask;
+    enum rules_kccgst kccgst_at_pos[_KCCGST_NUM_ENTRIES];
+    kccgst_index_t num_kccgst;
+    kccgst_mask_t defined_kccgst_mask;
 };
 
 enum mlvo_match_type {
@@ -248,9 +254,9 @@ enum wildcard_match_type {
 struct rule {
     struct sval mlvo_value_at_pos[_MLVO_NUM_ENTRIES];
     enum mlvo_match_type match_type_at_pos[_MLVO_NUM_ENTRIES];
-    unsigned int num_mlvo_values;
+    mlvo_index_t num_mlvo_values;
     struct sval kccgst_value_at_pos[_KCCGST_NUM_ENTRIES];
-    unsigned int num_kccgst_values;
+    kccgst_index_t num_kccgst_values;
     bool skip;
 };
 
@@ -343,7 +349,7 @@ matcher_new(struct xkb_context *ctx,
 
     m->ctx = ctx;
     m->rmlvo.model.sval.start = rmlvo->model;
-    m->rmlvo.model.sval.len = (unsigned int) strlen_safe(rmlvo->model);
+    m->rmlvo.model.sval.len = strlen_safe(rmlvo->model);
     m->rmlvo.layouts = split_comma_separated_mlvo(rmlvo->layout);
     m->rmlvo.variants = split_comma_separated_mlvo(rmlvo->variant);
     m->rmlvo.options = split_comma_separated_mlvo(rmlvo->options);
@@ -381,7 +387,7 @@ matcher_free(struct matcher *m)
         darray_free(group->elements);
     darray_free(m->pending_kccgst.buffer);
     darray_free(m->pending_kccgst.slices);
-    for (int i = 0; i < _KCCGST_NUM_ENTRIES; i++)
+    for (kccgst_index_t i = 0; i < (kccgst_index_t) _KCCGST_NUM_ENTRIES; i++)
         darray_free(m->kccgst[i]);
     darray_free(m->groups);
     free(m);
@@ -405,13 +411,13 @@ matcher_group_add_element(struct matcher *m, struct scanner *s,
 static bool
 read_rules_file(struct xkb_context *ctx,
                 struct matcher *matcher,
-                unsigned include_depth,
+                unsigned int include_depth,
                 FILE *file,
                 const char *path);
 
 static void
 matcher_include(struct matcher *m, struct scanner *parent_scanner,
-                unsigned include_depth,
+                unsigned int include_depth,
                 struct sval inc)
 {
     struct scanner s; /* parses the !include value */
@@ -423,7 +429,7 @@ matcher_include(struct matcher *m, struct scanner *parent_scanner,
 
     if (include_depth >= MAX_INCLUDE_DEPTH) {
         scanner_err(&s, XKB_LOG_MESSAGE_NO_ID,
-                    "maximum include depth (%d) exceeded; "
+                    "maximum include depth (%u) exceeded; "
                     "maybe there is an include loop?",
                     MAX_INCLUDE_DEPTH);
         return;
@@ -519,10 +525,10 @@ matcher_include(struct matcher *m, struct scanner *parent_scanner,
 static void
 matcher_mapping_start_new(struct matcher *m)
 {
-    for (unsigned i = 0; i < _MLVO_NUM_ENTRIES; i++)
-        m->mapping.mlvo_at_pos[i] = -1;
-    for (unsigned i = 0; i < _KCCGST_NUM_ENTRIES; i++)
-        m->mapping.kccgst_at_pos[i] = -1;
+    for (mlvo_index_t i = 0; i < (mlvo_index_t) _MLVO_NUM_ENTRIES; i++)
+        m->mapping.mlvo_at_pos[i] = _MLVO_NUM_ENTRIES;
+    for (kccgst_index_t i = 0; i < (kccgst_index_t) _KCCGST_NUM_ENTRIES; i++)
+        m->mapping.kccgst_at_pos[i] = _KCCGST_NUM_ENTRIES;
     m->mapping.has_layout_idx_range = false;
     m->mapping.layout_idx = m->mapping.variant_idx = XKB_LAYOUT_INVALID;
     m->mapping.num_mlvo = m->mapping.num_kccgst = 0;
@@ -630,7 +636,7 @@ matcher_mapping_set_mlvo(struct matcher *m, struct scanner *s,
         scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid mapping: \"%.*s\" is not a valid value here; "
                     "ignoring rule set",
-                    ident.len, ident.start);
+                    (unsigned int) ident.len, ident.start);
         m->mapping.active = false;
         return;
     }
@@ -639,7 +645,7 @@ matcher_mapping_set_mlvo(struct matcher *m, struct scanner *s,
         scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid mapping: \"%.*s\" appears twice on the same line; "
                     "ignoring rule set",
-                    mlvo_sval.len, mlvo_sval.start);
+                    (unsigned int) mlvo_sval.len, mlvo_sval.start);
         m->mapping.active = false;
         return;
     }
@@ -654,7 +660,7 @@ matcher_mapping_set_mlvo(struct matcher *m, struct scanner *s,
             scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                         "invalid mapping: \"%.*s\" may only be followed by a "
                         "valid group index; ignoring rule set",
-                        mlvo_sval.len, mlvo_sval.start);
+                        (unsigned int) mlvo_sval.len, mlvo_sval.start);
             m->mapping.active = false;
             return;
         }
@@ -669,7 +675,7 @@ matcher_mapping_set_mlvo(struct matcher *m, struct scanner *s,
             scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                         "invalid mapping: \"%.*s\" cannot be followed by a group "
                         "index; ignoring rule set",
-                        mlvo_sval.len, mlvo_sval.start);
+                        (unsigned int) mlvo_sval.len, mlvo_sval.start);
             m->mapping.active = false;
             return;
         }
@@ -688,7 +694,7 @@ matcher_mapping_set_mlvo(struct matcher *m, struct scanner *s,
     }
 
     m->mapping.mlvo_at_pos[m->mapping.num_mlvo] = mlvo;
-    m->mapping.defined_mlvo_mask |= 1u << mlvo;
+    m->mapping.defined_mlvo_mask |= (mlvo_mask_t) 1u << mlvo;
     m->mapping.num_mlvo++;
 }
 
@@ -749,7 +755,7 @@ matcher_mapping_set_kccgst(struct matcher *m, struct scanner *s, struct sval ide
         scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid mapping: \"%.*s\" is not a valid value here; "
                     "ignoring rule set",
-                    ident.len, ident.start);
+                    (unsigned int) ident.len, ident.start);
         m->mapping.active = false;
         return;
     }
@@ -758,13 +764,13 @@ matcher_mapping_set_kccgst(struct matcher *m, struct scanner *s, struct sval ide
         scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid mapping: \"%.*s\" appears twice on the same line; "
                     "ignoring rule set",
-                    kccgst_sval.len, kccgst_sval.start);
+                    (unsigned int) kccgst_sval.len, kccgst_sval.start);
         m->mapping.active = false;
         return;
     }
 
     m->mapping.kccgst_at_pos[m->mapping.num_kccgst] = kccgst;
-    m->mapping.defined_kccgst_mask |= 1u << kccgst;
+    m->mapping.defined_kccgst_mask |= (kccgst_mask_t) 1u << kccgst;
     m->mapping.num_kccgst++;
 }
 
@@ -853,7 +859,7 @@ matcher_rule_set_mlvo_common(struct matcher *m, struct scanner *s,
                              struct sval ident,
                              enum mlvo_match_type match_type)
 {
-    if (m->rule.num_mlvo_values + 1 > m->mapping.num_mlvo) {
+    if (m->rule.num_mlvo_values >= m->mapping.num_mlvo) {
         scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid rule: has more values than the mapping line; "
                     "ignoring rule");
@@ -891,7 +897,7 @@ static void
 matcher_rule_set_kccgst(struct matcher *m, struct scanner *s,
                         struct sval ident)
 {
-    if (m->rule.num_kccgst_values + 1 > m->mapping.num_kccgst) {
+    if (m->rule.num_kccgst_values >= m->mapping.num_kccgst) {
         scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid rule: has more values than the mapping line; "
                     "ignoring rule");
@@ -905,7 +911,7 @@ matcher_rule_set_kccgst(struct matcher *m, struct scanner *s,
 static bool
 match_group(struct matcher *m, struct sval group_name, struct sval to)
 {
-    struct group *group;
+    struct group *group = NULL;
     struct sval *element;
     bool found = false;
 
@@ -974,7 +980,7 @@ match_value_and_mark(struct matcher *m, struct sval val,
 static bool
 expand_rmlvo_in_kccgst_value(struct matcher *m, struct scanner *s,
                              struct sval value, xkb_layout_index_t layout_idx,
-                             darray_char *expanded, unsigned *i)
+                             darray_char *expanded, size_t *i)
 {
     const char *str = value.start;
 
@@ -1085,7 +1091,7 @@ expand_rmlvo_in_kccgst_value(struct matcher *m, struct scanner *s,
         darray_appends_nullterminate(*expanded, &pfx, 1);
     darray_appends_nullterminate(*expanded,
                                  expanded_value->sval.start,
-                                 expanded_value->sval.len);
+                                 (darray_size_t) expanded_value->sval.len);
     if (sfx != 0)
         darray_appends_nullterminate(*expanded, &sfx, 1);
     expanded_value->matched = true;
@@ -1107,7 +1113,7 @@ expand_qualifier_in_kccgst_value(
     struct matcher *m, struct scanner *s,
     struct sval value, darray_char *expanded,
     bool has_layout_idx_range, bool has_separator,
-    unsigned int prefix_idx, unsigned int *i)
+    darray_size_t prefix_idx, size_t *i)
 {
     const char *str = value.start;
 
@@ -1123,7 +1129,8 @@ expand_qualifier_in_kccgst_value(
         /* Check for more layouts (slow path) */
         if (darray_size(m->rmlvo.layouts) > 1) {
             char layout_index[MAX_LAYOUT_INDEX_STR_LENGTH + 1];
-            const size_t prefix_length = darray_size(*expanded) - prefix_idx - 1;
+            const darray_size_t prefix_length =
+                darray_size(*expanded) - prefix_idx - 1;
             xkb_layout_index_t l;
             for (l = 1;
                  l < MIN(XKB_MAX_GROUPS, darray_size(m->rmlvo.layouts));
@@ -1147,9 +1154,9 @@ expand_qualifier_in_kccgst_value(
 
 static inline void
 #ifdef _MSC_VER
-concat_kccgst(darray_char *into, size_t size, _In_reads_(size) const char* from)
+concat_kccgst(darray_char *into, darray_size_t size, _In_reads_(size) const char* from)
 #else
-concat_kccgst(darray_char *into, size_t size, const char from[static size])
+concat_kccgst(darray_char *into, darray_size_t size, const char from[static size])
 #endif
 {
     /*
@@ -1181,10 +1188,10 @@ append_expanded_kccgst_value(struct matcher *m, struct scanner *s,
 {
     const char *str = value.start;
     darray_char expanded = darray_new();
-    unsigned int last_item_idx = 0;
+    darray_size_t last_item_idx = 0;
     bool has_separator = false;
 
-    for (unsigned i = 0; i < value.len; ) {
+    for (size_t i = 0; i < value.len; ) {
         /* Check if that's a start of an expansion or qualifier */
         switch (str[i]) {
             /* Qualifier */
@@ -1241,7 +1248,7 @@ matcher_append_pending_kccgst(struct matcher *m)
      * Handle pending KcCGST values
      * See note: “Layout index ranges and merging KcCGST values”
      */
-    for (unsigned int i = 0; i < m->mapping.num_kccgst; i++) {
+    for (kccgst_index_t i = 0; i < m->mapping.num_kccgst; i++) {
         const enum rules_kccgst kccgst = m->mapping.kccgst_at_pos[i];
         /* For each relevant layout, append the relevant KcCGST values to
          * the output. */
@@ -1252,7 +1259,7 @@ matcher_append_pending_kccgst(struct matcher *m)
              * options. Process them sequentially. */
             register const struct kccgst_buffer* const buf = &m->pending_kccgst;
             size_t offset = 0;
-            for (unsigned k = 0; k < darray_size(buf->slices); k++) {
+            for (darray_size_t k = 0; k < darray_size(buf->slices); k++) {
                 register const struct kccgst_buffer_slice * const slice =
                     &darray_item(buf->slices, k);
                 if (slice->kccgst == kccgst && slice->layout == layout &&
@@ -1287,7 +1294,7 @@ matcher_rule_apply_if_matches(struct matcher *m, struct scanner *s)
     xkb_layout_mask_t candidate_layouts = m->mapping.layouts_candidates_mask;
     xkb_layout_index_t idx;
     /* Loop over MLVO pattern components */
-    for (unsigned i = 0; i < m->mapping.num_mlvo; i++) {
+    for (mlvo_index_t i = 0; i < m->mapping.num_mlvo; i++) {
         enum rules_mlvo mlvo = m->mapping.mlvo_at_pos[i];
         struct sval value = m->rule.mlvo_value_at_pos[i];
         enum mlvo_match_type match_type = m->rule.match_type_at_pos[i];
@@ -1360,7 +1367,7 @@ matcher_rule_apply_if_matches(struct matcher *m, struct scanner *s)
              idx++)
         {
             if (candidate_layouts & (UINT32_C(1) << idx)) {
-                for (unsigned i = 0; i < m->mapping.num_kccgst; i++) {
+                for (kccgst_index_t i = 0; i < m->mapping.num_kccgst; i++) {
                     const enum rules_kccgst kccgst = m->mapping.kccgst_at_pos[i];
                     const struct sval value = m->rule.kccgst_value_at_pos[i];
                     /*
@@ -1402,7 +1409,8 @@ matcher_rule_apply_if_matches(struct matcher *m, struct scanner *s)
                      */
                     register struct kccgst_buffer * const buf =
                         &m->pending_kccgst;
-                    const size_t prev_buffer_length = darray_size(buf->buffer);
+                    const darray_size_t prev_buffer_length =
+                        darray_size(buf->buffer);
                     append_expanded_kccgst_value(m, s, false, &buf->buffer,
                                                  value, idx);
                     const uint32_t length = (uint32_t) (darray_size(buf->buffer)
@@ -1418,7 +1426,7 @@ matcher_rule_apply_if_matches(struct matcher *m, struct scanner *s)
         }
     } else {
         /* Numeric index or no index */
-        for (unsigned i = 0; i < m->mapping.num_kccgst; i++) {
+        for (kccgst_index_t i = 0; i < m->mapping.num_kccgst; i++) {
             enum rules_kccgst kccgst = m->mapping.kccgst_at_pos[i];
             struct sval value = m->rule.kccgst_value_at_pos[i];
             append_expanded_kccgst_value(m, s, true, &m->kccgst[kccgst], value,
@@ -1444,7 +1452,7 @@ gettok(struct matcher *m, struct scanner *s)
 
 static bool
 matcher_match(struct matcher *m, struct scanner *s,
-              unsigned include_depth,
+              unsigned int include_depth,
               const char *string, size_t len,
               const char *file_name)
 {
@@ -1639,7 +1647,7 @@ error:
 static bool
 read_rules_file(struct xkb_context *ctx,
                 struct matcher *matcher,
-                unsigned include_depth,
+                unsigned int include_depth,
                 FILE *file,
                 const char *path)
 {
@@ -1721,22 +1729,22 @@ xkb_components_from_rules(struct xkb_context *ctx,
     if (!mval->matched && mval->sval.len > 0)
         log_err(matcher->ctx, XKB_ERROR_CANNOT_RESOLVE_RMLVO,
                 "Unrecognized RMLVO model \"%.*s\" was ignored\n",
-                mval->sval.len, mval->sval.start);
+                (unsigned int) mval->sval.len, mval->sval.start);
     darray_foreach(mval, matcher->rmlvo.layouts)
         if (!mval->matched && mval->sval.len > 0)
             log_err(matcher->ctx, XKB_ERROR_CANNOT_RESOLVE_RMLVO,
                     "Unrecognized RMLVO layout \"%.*s\" was ignored\n",
-                    mval->sval.len, mval->sval.start);
+                    (unsigned int) mval->sval.len, mval->sval.start);
     darray_foreach(mval, matcher->rmlvo.variants)
         if (!mval->matched && mval->sval.len > 0)
             log_err(matcher->ctx, XKB_ERROR_CANNOT_RESOLVE_RMLVO,
                     "Unrecognized RMLVO variant \"%.*s\" was ignored\n",
-                    mval->sval.len, mval->sval.start);
+                    (unsigned int) mval->sval.len, mval->sval.start);
     darray_foreach(mval, matcher->rmlvo.options)
         if (!mval->matched && mval->sval.len > 0)
             log_err(matcher->ctx, XKB_ERROR_CANNOT_RESOLVE_RMLVO,
                     "Unrecognized RMLVO option \"%.*s\" was ignored\n",
-                    mval->sval.len, mval->sval.start);
+                    (unsigned int) mval->sval.len, mval->sval.start);
 
     /* Set the number of explicit layouts */
     if (out->symbols != NULL && explicit_layouts != NULL) {
