@@ -14,6 +14,7 @@
 
 #include "config.h"
 
+#include "messages-codes.h"
 #include "xkbcommon/xkbcommon.h"
 #include "xkbcommon/xkbcommon-keysyms.h"
 
@@ -217,9 +218,6 @@ static bool
 MergeGroups(SymbolsInfo *info, GroupInfo *into, GroupInfo *from, bool clobber,
             bool report, xkb_layout_index_t group, xkb_atom_t key_name)
 {
-    xkb_level_index_t i, levels_in_both;
-    struct xkb_level *level;
-
     /* First find the type of the merged group. */
     if (into->type != from->type) {
         if (from->type == XKB_ATOM_NONE) {
@@ -234,7 +232,7 @@ MergeGroups(SymbolsInfo *info, GroupInfo *into, GroupInfo *from, bool clobber,
 
             if (report) {
                 log_warn(info->ctx, XKB_WARNING_CONFLICTING_KEY_TYPE_MERGING_GROUPS,
-                         "Multiple definitions for group %d type of key %s; "
+                         "Multiple definitions for group %"PRIu32" type of key %s; "
                          "Using %s, ignoring %s\n",
                          group + 1, KeyNameText(info->ctx, key_name),
                          xkb_atom_text(info->ctx, use),
@@ -261,12 +259,13 @@ MergeGroups(SymbolsInfo *info, GroupInfo *into, GroupInfo *from, bool clobber,
     }
 
     /* Merge the actions and syms. */
-    levels_in_both = MIN(darray_size(into->levels), darray_size(from->levels));
-    unsigned int fromKeysymsCount = 0;
-    unsigned int fromActionsCount = 0;
-    for (i = 0; i < levels_in_both; i++) {
-        struct xkb_level *intoLevel = &darray_item(into->levels, i);
-        struct xkb_level *fromLevel = &darray_item(from->levels, i);
+    const darray_size_t levels_in_both =
+        MIN(darray_size(into->levels), darray_size(from->levels));
+    darray_size_t fromKeysymsCount = 0;
+    darray_size_t fromActionsCount = 0;
+    for (darray_size_t i = 0; i < levels_in_both; i++) {
+        struct xkb_level* const intoLevel = &darray_item(into->levels, i);
+        struct xkb_level* const fromLevel = &darray_item(from->levels, i);
 
         const bool fromHasNoKeysym = fromLevel->num_syms == 0;
         const bool fromHasNoAction = fromLevel->num_actions == 0;
@@ -294,8 +293,8 @@ MergeGroups(SymbolsInfo *info, GroupInfo *into, GroupInfo *from, bool clobber,
             /* Incompatible keysyms */
             if (report && !(intoHasNoKeysym || fromHasNoKeysym)) {
                 log_warn(info->ctx, XKB_WARNING_CONFLICTING_KEY_SYMBOL,
-                            "Multiple symbols for level %d/group %u on key %s; "
-                            "Using %s, ignoring %s\n",
+                            "Multiple symbols for level %u/group "
+                            "%"PRIu32" on key %s; Using %s, ignoring %s\n",
                             i + 1, group + 1, KeyNameText(info->ctx, key_name),
                             (clobber ? "from" : "to"),
                             (clobber ? "to" : "from"));
@@ -349,8 +348,8 @@ MergeGroups(SymbolsInfo *info, GroupInfo *into, GroupInfo *from, bool clobber,
                 if (unlikely(intoLevel->num_actions > 1)) {
                     log_warn(
                         info->ctx, XKB_WARNING_CONFLICTING_KEY_ACTION,
-                        "Multiple actions for level %d/group %u on key %s; "
-                        "%s\n",
+                        "Multiple actions for level %u/group %"PRIu32" "
+                        "on key %s; %s\n",
                         i + 1, group + 1, KeyNameText(info->ctx, key_name),
                         clobber ? "Using from, ignoring to"
                                 : "Using to, ignoring from");
@@ -364,8 +363,8 @@ MergeGroups(SymbolsInfo *info, GroupInfo *into, GroupInfo *from, bool clobber,
                         : &fromLevel->a.action;
 
                     log_warn(info->ctx, XKB_WARNING_CONFLICTING_KEY_ACTION,
-                                "Multiple actions for level %d/group %u "
-                                "on key %s; Using %s, ignoring %s\n",
+                                "Multiple actions for level %u/group "
+                                "%"PRIu32" on key %s; Using %s, ignoring %s\n",
                                 i + 1, group + 1,
                                 KeyNameText(info->ctx, key_name),
                                 ActionTypeText(use->type),
@@ -415,6 +414,7 @@ MergeGroups(SymbolsInfo *info, GroupInfo *into, GroupInfo *from, bool clobber,
         }
     }
     /* If @from has extra levels, get them as well. */
+    struct xkb_level *level;
     darray_foreach_from(level, from->levels, levels_in_both) {
         darray_append(into->levels, *level);
         /* We may have stolen keysyms or actions arrays:
@@ -686,8 +686,8 @@ HandleIncludeSymbols(SymbolsInfo *info, IncludeStmt *include)
             next_incl.explicit_group = atoi(stmt->modifier) - 1;
             if (next_incl.explicit_group >= XKB_MAX_GROUPS) {
                 log_err(info->ctx, XKB_ERROR_UNSUPPORTED_GROUP_INDEX,
-                        "Cannot set explicit group to %d - must be between 1..%d; "
-                        "Ignoring group number\n",
+                        "Cannot set explicit group to %"PRIu32" - "
+                        "must be between 1..%u; Ignoring group number\n",
                         next_incl.explicit_group + 1, XKB_MAX_GROUPS);
                 next_incl.explicit_group = info->explicit_group;
             }
@@ -785,7 +785,7 @@ AddSymbolsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
     if (value->common.type != STMT_EXPR_KEYSYM_LIST) {
         log_err(info->ctx, XKB_ERROR_WRONG_FIELD_TYPE,
                 "Expected a list of symbols, found %s; "
-                "Ignoring symbols for group %u of %s\n",
+                "Ignoring symbols for group %"PRIu32" of %s\n",
                 stmt_type_to_string(value->common.type), ndx + 1,
                 KeyInfoText(info, keyi));
         return false;
@@ -793,7 +793,7 @@ AddSymbolsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
 
     if (groupi->defined & GROUP_FIELD_SYMS) {
         log_err(info->ctx, XKB_ERROR_CONFLICTING_KEY_SYMBOLS_ENTRY,
-                "Symbols for key %s, group %u already defined; "
+                "Symbols for key %s, group %"PRIu32" already defined; "
                 "Ignoring duplicate definition\n",
                 KeyInfoText(info, keyi), ndx + 1);
         return false;
@@ -827,8 +827,8 @@ AddSymbolsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
 
         if (unlikely(darray_size(keysymList->syms) > MAX_KEYSYMS_PER_LEVEL)) {
             log_err(info->ctx, XKB_LOG_MESSAGE_NO_ID,
-                    "Key %s has too many keysyms for group %u, level %u; "
-                    "expected max %u, got: %u\n",
+                    "Key %s has too many keysyms for group %"PRIu32", "
+                    "level %"PRIu32"; expected max %u, got: %u\n",
                     KeyInfoText(info, keyi), ndx + 1, level + 1,
                     MAX_KEYSYMS_PER_LEVEL, darray_size(keysymList->syms));
             return false;
@@ -876,14 +876,14 @@ AddActionsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
     if (value->common.type != STMT_EXPR_ACTION_LIST) {
         log_wsgo(info->ctx, XKB_ERROR_INVALID_EXPRESSION_TYPE,
                  "Bad expression type (%d) for action list value; "
-                 "Ignoring actions for group %u of %s\n",
+                 "Ignoring actions for group %"PRIu32" of %s\n",
                  value->common.type, ndx, KeyInfoText(info, keyi));
         return false;
     }
 
     if (groupi->defined & GROUP_FIELD_ACTS) {
         log_wsgo(info->ctx, XKB_WARNING_CONFLICTING_KEY_ACTION,
-                 "Actions for key %s, group %u already defined\n",
+                 "Actions for key %s, group %"PRIu32" already defined\n",
                  KeyInfoText(info, keyi), ndx);
         return false;
     }
@@ -914,8 +914,8 @@ AddActionsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
 
         if (unlikely(num_actions > MAX_ACTIONS_PER_LEVEL)) {
             log_err(info->ctx, XKB_LOG_MESSAGE_NO_ID,
-                    "Key %s has too many actions for group %u, level %u; "
-                    "expected max %u, got: %u\n",
+                    "Key %s has too many actions for group %"PRIu32", "
+                    "level %"PRIu32"; expected max %u, got: %u\n",
                     KeyInfoText(info, keyi), ndx + 1, level + 1,
                     MAX_ACTIONS_PER_LEVEL, num_actions);
             return false;
@@ -923,16 +923,18 @@ AddActionsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
 
         /* Parse actions and add only defined actions */
         darray(union xkb_action) actions = darray_new();
-        unsigned int act_index = 0;
         for (ExprDef *act = actionList->actions;
-             act; act = (ExprDef *) act->common.next, act_index++) {
+             act; act = (ExprDef *) act->common.next) {
             union xkb_action toAct = { 0 };
             if (!HandleActionDef(info->ctx, &info->default_actions, &info->mods,
-                                 act, &toAct))
+                                 act, &toAct)) {
                 log_err(info->ctx, XKB_ERROR_INVALID_VALUE,
                         "Illegal action definition for %s; "
-                        "Action for group %u/level %u ignored\n",
+                        "Action for group %"PRIu32"/level %"PRIu32" ignored\n",
                         KeyInfoText(info, keyi), ndx + 1, level + 1);
+                /* Ensure action type is reset */
+                toAct.type = ACTION_TYPE_NONE;
+            }
             if (toAct.type == ACTION_TYPE_NONE) {
                 /* Drop action */
             } else if (likely(num_actions == 1)) {
@@ -1176,7 +1178,7 @@ SetGroupName(SymbolsInfo *info, ExprDef *arrayNdx, ExprDef *value,
     if (!ExprResolveString(info->ctx, value, &name)) {
         log_err(info->ctx, XKB_ERROR_WRONG_FIELD_TYPE,
                 "Group name must be a string; "
-                "Illegal name for group %d ignored\n", group);
+                "Illegal name for group %"PRIu32" ignored\n", group);
         return false;
     }
 
@@ -1190,8 +1192,8 @@ SetGroupName(SymbolsInfo *info, ExprDef *arrayNdx, ExprDef *value,
     else {
         log_warn(info->ctx, XKB_WARNING_NON_BASE_GROUP_NAME,
                  "An explicit group was specified for the '%s' map, "
-                 "but it provides a name for a group other than Group1 (%d); "
-                 "Ignoring group name '%s'\n",
+                 "but it provides a name for a group other than Group1 "
+                 "(%"PRIu32"); Ignoring group name '%s'\n",
                  info->name, group,
                  xkb_atom_text(info->ctx, name));
         return false;
@@ -1269,10 +1271,14 @@ HandleGlobalVar(SymbolsInfo *info, VarDef *stmt)
                 "Ignoring \"allownone\" specification\n");
         ret = true;
     }
-    else {
+    else if (elem) {
         ret = SetDefaultActionField(info->ctx, &info->default_actions,
                                     &info->mods, elem, field, arrayNdx,
                                     stmt->value, stmt->merge);
+    } else {
+        log_err(info->ctx, XKB_ERROR_UNKNOWN_DEFAULT_FIELD,
+                "Default defined for unknown field \"%s\"; Ignored\n", field);
+        return false;
     }
 
     return ret;
@@ -1346,7 +1352,7 @@ SetExplicitGroup(SymbolsInfo *info, KeyInfo *keyi)
 
     if (warn) {
         log_warn(info->ctx, XKB_WARNING_MULTIPLE_GROUPS_AT_ONCE,
-                 "For the map %s the explicit group %u is specified, "
+                 "For the map %s the explicit group %"PRIu32" is specified, "
                  "but key %s has more than one group defined; "
                  "All groups except first one will be ignored\n",
                  info->name, info->explicit_group + 1, KeyInfoText(info, keyi));
@@ -1500,23 +1506,23 @@ HandleSymbolsFile(SymbolsInfo *info, XkbFile *file)
 static struct xkb_key *
 FindKeyForSymbol(struct xkb_keymap *keymap, xkb_keysym_t sym)
 {
-    struct xkb_key *key;
-    xkb_layout_index_t group;
-    bool got_one_group, got_one_level;
-
-    group = 0;
+    bool got_one_group;
+    xkb_layout_index_t group = 0;
     do {
         xkb_level_index_t level = 0;
         got_one_group = false;
+        bool got_one_level;
         do {
             got_one_level = false;
+            struct xkb_key *key;
             xkb_keys_foreach(key, keymap) {
                 if (group < key->num_groups &&
                     level < XkbKeyNumLevels(key, group)) {
                     got_one_group = got_one_level = true;
-                    unsigned int num_syms = key->groups[group].levels[level].num_syms;
+                    const xkb_keysym_count_t num_syms =
+                        key->groups[group].levels[level].num_syms;
                     if (num_syms > 1) {
-                        for (unsigned int k = 0; k < num_syms; k++) {
+                        for (xkb_keysym_count_t k = 0; k < num_syms; k++) {
                             if (key->groups[group].levels[level].s.syms[k] == sym)
                                 return key;
                         }
@@ -1603,7 +1609,6 @@ static const struct xkb_key_type *
 FindTypeForGroup(struct xkb_keymap *keymap, KeyInfo *keyi,
                  xkb_layout_index_t group, bool *explicit_type)
 {
-    unsigned int i;
     GroupInfo *groupi = &darray_item(keyi->groups, group);
     xkb_atom_t type_name = groupi->type;
 
@@ -1623,19 +1628,20 @@ FindTypeForGroup(struct xkb_keymap *keymap, KeyInfo *keyi,
     if (type_name == XKB_ATOM_NONE) {
         log_warn(keymap->ctx, XKB_WARNING_CANNOT_INFER_KEY_TYPE,
                  "Couldn't find an automatic type for key '%s' "
-                 "group %d with %lu levels; Using the default type\n",
+                 "group %"PRIu32" with %u levels; Using the default type\n",
                  KeyNameText(keymap->ctx, keyi->name), group + 1,
-                 (unsigned long) darray_size(groupi->levels));
+                 darray_size(groupi->levels));
         goto use_default;
     }
 
+    darray_size_t i;
     for (i = 0; i < keymap->num_types; i++)
         if (keymap->types[i].name == type_name)
             break;
 
     if (i >= keymap->num_types) {
         log_warn(keymap->ctx, XKB_WARNING_UNDEFINED_KEY_TYPE,
-                 "The type \"%s\" for key '%s' group %d "
+                 "The type \"%s\" for key '%s' group %"PRIu32" "
                  "was not previously defined; Using the default type\n",
                  xkb_atom_text(keymap->ctx, type_name),
                  KeyNameText(keymap->ctx, keyi->name), group + 1);
@@ -1723,7 +1729,7 @@ CopySymbolsDefToKeymap(struct xkb_keymap *keymap, SymbolsInfo *info,
             struct xkb_level *leveli;
 
             log_vrb(info->ctx, 1, XKB_WARNING_EXTRA_SYMBOLS_IGNORED,
-                    "Type \"%s\" has %d levels, but %s has %u levels; "
+                    "Type \"%s\" has %"PRIu32" levels, but %s has %u levels; "
                     "Ignoring extra symbols\n",
                     xkb_atom_text(keymap->ctx, type->name), type->num_levels,
                     KeyInfoText(info, keyi),
@@ -1864,8 +1870,8 @@ CopySymbolsToKeymap(struct xkb_keymap *keymap, SymbolsInfo *info)
 
     keymap->mods = info->mods;
 
-    darray_steal(info->group_names,
-                 &keymap->group_names, &keymap->num_group_names);
+    keymap->num_group_names = (xkb_layout_index_t)darray_size(info->group_names);
+    darray_steal(info->group_names, &keymap->group_names, NULL);
 
     darray_foreach(keyi, info->keys)
         if (!CopySymbolsDefToKeymap(keymap, info, keyi))
